#include <Adafruit_NeoPixel\Adafruit_NeoPixel.h>

#include <Wire.h>
#include <SPI.h>
#include <SparkFunLSM9DS1.h>

#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <ESP8266WebServer.h>
#include <ESP8266mDNS.h>
#include <WiFiUdp.h>
#include <ArduinoOTA.h>
#include <FS.h>
#include "ConfigMgr.h"
#include "DNSServer.h"

#define WIFI


#ifdef WIFI
ESP8266WebServer server(80);
DNSServer * dns = nullptr;
#endif

#define PIN 2
Adafruit_NeoPixel strip = Adafruit_NeoPixel(7, PIN, NEO_GRB + NEO_KHZ800);

#define FACTORY_RESET_PIN 16

LSM9DS1 imu;
#define LSM9DS1_M	0x1E // Would be 0x1C if SDO_M is LOW
#define LSM9DS1_AG	0x6B // Would be 0x6A if SDO_AG is LOW

#ifdef WIFI
void handleLog() {
	server.send(200, "text/plain", logBuffer);
}

void HandleConfig() {
	if (server.hasArg("action") &&
		server.arg("action") == "config") {
		if (server.hasArg("ip")) ConfigMgr::Instance()->GetConfig().IP.fromString(server.arg("ip").c_str());
		LOGF("Converting ip %s to uint %x\n", server.arg("ip").c_str(), ConfigMgr::Instance()->GetConfig().IP.toString().c_str());
		if (server.hasArg("gateway")) ConfigMgr::Instance()->GetConfig().Gateway.fromString(server.arg("gateway").c_str());
		if (server.hasArg("subnet")) ConfigMgr::Instance()->GetConfig().Subnet.fromString(server.arg("subnet").c_str());
		if (server.hasArg("id")) ConfigMgr::Instance()->GetConfig().Id = atoi(server.arg("id").c_str());
		if (server.hasArg("ssid")) strcpy(ConfigMgr::Instance()->GetConfig().ssid, server.arg("ssid").c_str());
		if (server.hasArg("password")) strcpy(ConfigMgr::Instance()->GetConfig().password, server.arg("password").c_str());
		ConfigMgr::Instance()->Write();
	}

	char buff[2048];
	char ipStr[20] = "";
	char subnetStr[20] = "";
	char gatewayStr[20] = "";
	char idStr[20];
	if (ConfigMgr::Instance()->HasConfig) {
		itoa(ConfigMgr::Instance()->GetConfig().Id, idStr, 10);
		sprintf(buff, configFormat,
			ConfigMgr::Instance()->GetConfig().IP.toString().c_str(),
			ConfigMgr::Instance()->GetConfig().Gateway.toString().c_str(),
			ConfigMgr::Instance()->GetConfig().Subnet.toString().c_str(),
			idStr,
			ConfigMgr::Instance()->GetConfig().ssid,
			ConfigMgr::Instance()->GetConfig().password
		);
	}
	else {
		sprintf(buff, configFormat, "", "", "", "", "", "", "");

	}
	server.send(200, "text/html", buff);
}
void HandleCommand() {
	if (server.hasArg("command")) {
		Serial.println(server.arg("command").c_str());
	}

	server.send(204, "", "");
}

String getContentType(String filename) {
	if (server.hasArg("download")) return "application/octet-stream";
	else if (filename.endsWith(".htm")) return "text/html";
	else if (filename.endsWith(".html")) return "text/html";
	else if (filename.endsWith(".css")) return "text/css";
	else if (filename.endsWith(".js")) return "application/javascript";
	else if (filename.endsWith(".png")) return "image/png";
	else if (filename.endsWith(".gif")) return "image/gif";
	else if (filename.endsWith(".jpg")) return "image/jpeg";
	else if (filename.endsWith(".ico")) return "image/x-icon";
	else if (filename.endsWith(".xml")) return "text/xml";
	else if (filename.endsWith(".pdf")) return "application/x-pdf";
	else if (filename.endsWith(".zip")) return "application/x-zip";
	else if (filename.endsWith(".gz")) return "application/x-gzip";
	return "text/plain";
}

bool handleFileRead(String path) {
	LOGF("Reading file %s\n", path.c_str());
	//DBG_OUTPUT_PORT.println("handleFileRead: " + path);
	if (path.endsWith("/")) path += "index.htm";
	String contentType = getContentType(path);
	String pathWithGz = path + ".gz";
	if (SPIFFS.exists(pathWithGz) || SPIFFS.exists(path)) {
		if (SPIFFS.exists(pathWithGz))
			path += ".gz";
		File file = SPIFFS.open(path, "r");
		server.sendHeader("Cache-Control", "max-age=315360000");
		server.sendHeader("Expires", "Sun, 19 Jul 2020 18:06:32 GMT");

		size_t sent = server.streamFile(file, contentType);
		file.close();
		return true;
	}
	return false;
}

void listFiles() {
	Dir dir = SPIFFS.openDir("/");
	LOG("Files list:\n");
	while (dir.next()) {
		LOGF("File: %s\n", dir.fileName().c_str());
	}
}

#endif

void setup()
{
	Serial.begin(115200);

#ifdef WIFI
	SPIFFS.begin();
	LOG("SETUP\n");
	SPIFFS.begin();

	//  pinMode(led, OUTPUT);
	//  digitalWrite(led, 0);

	Serial.println("Hello!");
	pinMode(FACTORY_RESET_PIN, INPUT_PULLUP);
	delay(2);

	if (digitalRead(FACTORY_RESET_PIN) == LOW) {
		LOG("FACTORY RESET!\n");
		Serial.print("FACTORY RESET!\n");
		ConfigMgr::Instance()->FactoryReset();
	}

	if (ConfigMgr::Instance()->HasConfig) {
		Serial.println("Has config");;
		LOG("Begining wifi from config\n");
		Serial.println("Log print");
		WiFi.mode(WIFI_STA);
		WiFi.begin(ConfigMgr::Instance()->GetConfig().ssid,
			ConfigMgr::Instance()->GetConfig().password);
		Serial.println("Wifi.begin");

		// Wait for connection
		while (WiFi.status() != WL_CONNECTED) {
			delay(500);
			LOG(".");
		}
		Serial.printf("Wifi connected to %s\n", ConfigMgr::Instance()->GetConfig().ssid);
		WiFi.config(ConfigMgr::Instance()->GetConfig().IP,
			ConfigMgr::Instance()->GetConfig().Gateway,
			ConfigMgr::Instance()->GetConfig().Subnet);
		Serial.println("Wifi configured");

		Serial.printf("\nConnected to %s. Address: %s\n",
			ConfigMgr::Instance()->GetConfig().ssid,
			WiFi.localIP().toString().c_str());
	}
	else {
		LOG("Starting as AP\n");
		Serial.print("Starting as AP\n");
		WiFi.mode(WIFI_AP);
		WiFi.softAPConfig(
			IPAddress(10, 10, 10, 1),
			IPAddress(10, 10, 10, 1),
			IPAddress(255, 255, 255, 0));
		WiFi.softAP("Spacetime Poi");
		dns = new DNSServer();
		dns->start(53, "*", IPAddress(10, 10, 10, 1));

	}

	if (MDNS.begin("touchyblinky")) {
		LOG("MDNS responder started\n");
	}

	listFiles();
	server.on("/", []() {
		server.sendHeader("Location", "/index.html");
		server.send(303, "", "");
	});
	server.on("/control", HandleCommand);
	server.on("/config", HandleConfig);
	server.on("/log", handleLog);

	//called when the url is not defined here
	//use it to load content from SPIFFS
	server.onNotFound([]() {
		if (!handleFileRead(server.uri()))
			server.send(404, "text/plain", "FileNotFound");
	});

	server.begin();

	LOG("HTTP server started\n");
	ArduinoOTA.onStart([]() { LOG("Start"); });
	ArduinoOTA.onEnd([]() { LOG("\nEnd\n"); });
	ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) {
		LOGF("Progress: %u%%\r", (progress / (total / 100)));
	});

	ArduinoOTA.onError([](ota_error_t error) {
		LOGF("Error[%u]: ", error);
		if (error == OTA_AUTH_ERROR) LOG("Auth Failed\n");
		else if (error == OTA_BEGIN_ERROR) LOG("Begin Failed\n");
		else if (error == OTA_CONNECT_ERROR) LOG("Connect Failed\n");
		else if (error == OTA_RECEIVE_ERROR) LOG("Receive Failed\n");
		else if (error == OTA_END_ERROR) LOG("End Failed\n");
	});
	ArduinoOTA.begin();

#endif

	strip.begin();
	strip.show(); // Initialize all pixels to 'off'

	imu.settings.device.commInterface = IMU_MODE_I2C;
	imu.settings.device.mAddress = LSM9DS1_M;
	imu.settings.device.agAddress = LSM9DS1_AG;
	imu.settings.gyro.scale = M_SCALE_16GS;
	imu.settings.accel.scale = A_SCALE_16G;
	// The above lines will only take effect AFTER calling
	// imu.begin(), which verifies communication with the IMU
	// and turns it on.
	if (!imu.begin())
	{
		Serial.println("Failed to communicate with LSM9DS1.");
		Serial.println("Double-check wiring.");
		Serial.println("Default settings in this sketch will " \
			"work for an out of the box LSM9DS1 " \
			"Breakout, but may need to be modified " \
			"if the board jumpers are.");
		while (1)
			;
	}
}

int frame = 0;

int mode = 2;


int candle() {
	uint16_t color = (random(50) < 6) ? 0 : 255;
	for (int i = 0; i < 7; i++) {
		color = (color + (random(50) < 6) ? 0 : 255) / 2;
		strip.setPixelColor(i, color, 0, 0);
	}
	return 50;
}

int blue_move() {
	for (int i = 0; i < 7; i++) {
		strip.setPixelColor(i, 0, 0, (frame%7 == i) ? 255 : 0);
	}
	return 10;
}

int fade() {
	for (int i = 0; i < 7; i++) {
		int c = (frame + i * (255 / 7)) % 255;
		if (c > 127) c = 255 - c;
		strip.setPixelColor(i, c, 0, 0);
	}

	return 4;
}

int special_1() {
	auto az = imu.calcAccel(imu.az);
	auto ax = imu.calcAccel(imu.ax);

	if (az > -1.10) {
		return candle();
	}
	else if (az > -1.60) {
		for (int i = 0; i < 7; i++) {
			if (ax < 0) {
				strip.setPixelColor(i, (frame % 7 == i) ? 255 : 0, 0, 0);;
			}
			else {
				uint8_t c = (frame % 7 == i) ? 127 : 0;
				strip.setPixelColor(i, c, c/2, 0);;
			}
		}
		return 175 + (az * 100);
	}
	else {
		for (int i = 0; i < 7; i++) {
			if (ax < 0) {
				strip.setPixelColor(i, 0,0, (frame % 7 == i) ? 255 : 0);;
			}
			else {
				uint8_t c = (frame % 7 == i) ? 80 : 0;
				strip.setPixelColor(i, c, c, c);
			}
		}
		return 10;
	}
}

void printAccel()
{
	LOGF("A: x [%d] y [%d] z [%d]\n", (int)(imu.calcAccel(imu.ax) * 100), (int)(imu.calcAccel(imu.ay) * 100), (int)(imu.calcAccel(imu.az) * 100));
}
void printGyro()
{
	LOGF("G: x [%d] y [%d] z [%d]\n", (int)(imu.calcAccel(imu.gx) * 100), (int)(imu.calcAccel(imu.gy) * 100), (int)(imu.calcAccel(imu.gz) * 100));
}
void printMag()
{
	LOGF("M: x [%d] y [%d] z [%d]\n", (int)(imu.calcAccel(imu.mx) * 100), (int)(imu.calcAccel(imu.my) * 100), (int)(imu.calcAccel(imu.mz) * 100));
}

void loop()
{

	int delay = 10;
	switch (mode) {
	case 0: delay = candle(); break;
	case 1: delay = blue_move(); break;
	case 2: delay = special_1(); break;
	case 3: delay = fade(); break; 
	}
	strip.show();
	frame++;
	auto start = millis();
	while (millis() - start < delay) {
		if (imu.accelAvailable()) {
			imu.readAccel();
			printAccel();
		}
		if (false && imu.gyroAvailable())
		{
			imu.readGyro();
			printGyro();
		}
		if (false && imu.magAvailable())
		{
			imu.readMag();
			printMag();
		}

#ifdef WIFI
		if (dns) dns->processNextRequest();
		server.handleClient();
		ArduinoOTA.handle();
#endif
	}
}
