#pragma once

#include <EEPROM.h>
//#include <EEPROM\EEPROM.h>
#include <FS.h>
#include "LOG.h"

struct CONFIG {
	IPAddress IP;
	IPAddress Gateway;
	IPAddress Subnet;
	uint8_t Id;
	char ssid[30];
	char password[30];
};

class ConfigMgr {
public:
	CONFIG& GetConfig() {
		return m_config;
	}
	bool HasConfig;
	void Read() {
		LOG("Reading config\n");
		if (!SPIFFS.exists("/config.cfg")) {
			LOG("No config file!\n");  
			return;
		}
		auto file = SPIFFS.open("/config.cfg", "r");
		if (file.available() < sizeof(MAGIC) + sizeof(m_config)) {
			LOG("Config file too short\n");
			file.close();
			return;
		}
		LOG("Enough data\n");
		uint64_t magic;
		file.readBytes((char*)&magic, sizeof(magic));
		if (magic != MAGIC) {
			LOGF("No magic! expected: %ull got : %ull\n", MAGIC, magic);
			HasConfig = false;
			file.close();
			return;
		}
		file.readBytes((char*)&m_config, sizeof(m_config));
		file.close();
		HasConfig = true;
		LOG("Loaded config file!\n");
	}
	void Write() {
		auto fsUploadFile = SPIFFS.open("/config.cfg", "w");
		fsUploadFile.write((uint8_t*)&MAGIC, sizeof(MAGIC));
		fsUploadFile.write((uint8_t*)&m_config, sizeof(m_config));
		fsUploadFile.flush();
		fsUploadFile.close();
		LOG("Wrote config!\n");
		Read();
		HasConfig = true;
	}
	void FactoryReset() {
		SPIFFS.remove("/config.cfg");
		HasConfig = false;
	}
private:
	ConfigMgr() {
		read = false;
	}
	static ConfigMgr s_instance;
	CONFIG m_config;
	bool read;
	uint64_t MAGIC = *(uint64_t*)"TCBLCNFG";
public:
	static ConfigMgr* Instance() {
		if (!s_instance.read) {
			LOG("FIRST READ\n");
			//SPIFFS.begin();
			delay(10);
			s_instance.Read();
			s_instance.read = true;
		}
		return &s_instance;
	}
};
ConfigMgr ConfigMgr::s_instance;
const char* configFormat = R"====(
<!DOCTYPE html>

<html>
<head>
<link rel="icon" href="favicon.ico?v=2" />
</head>

<body>
<form action='config'>
	<table border="1">
			<input type='hidden' name='action' value='config' >
			<tr><td>IP:</td><td><input class="form-control" data-fv-field="ip" name="ip" type="text" value="%s"></input></td></tr>
			<tr><td>Gateway:</td><td><input class="form-control" data-fv-field="ip" name="gateway" type="text" value="%s"></input></td></tr>
			<tr><td>Subnet: </td><td><input class="form-control" data-fv-field="ip" name="subnet" type="text" value="%s"></input></td></tr>
			<tr><td>ID (1-254): </td><td><input class="form-control" name="id" type="text" value="%s"></input></td></tr>
			<tr><td>SSID: </td><td><input class="form-control" name="ssid" type="text" value="%s"></input></td></tr>
			<tr><td>Password: </td><td><input class="form-control" name="password" type="text" value="%s"></input></td></tr>
	</table>
    <input type='submit' value='Submit'><br>
</form>
</body>

)====";
